from snowflake.ingest.utils import SecurityManager


def test_same_token(test_util):
    """
    Tests that accounts are parsed correctly
    """
    expected_account = 'TESTACCOUNT'

    actual_account = 'testaccount'
    user = 'testuser'
    private_key = ''
    sec_manager = SecurityManager(actual_account, user, private_key)
    assert sec_manager.get_account() == expected_account

    actual_account = 'testaccount.something'
    sec_manager = SecurityManager(actual_account, user, private_key)
    assert sec_manager.get_account() == expected_account



# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
"""
test_tokens.py - This defines a series of tests to ascertain that we are
capable of renewing JWT tokens
"""

from snowflake.ingest.utils import SecurityManager
from snowflake.ingest.error import IngestClientError
from snowflake.ingest.errorcode import ERR_INVALID_PRIVATE_KEY
from datetime import timedelta
from time import sleep
import os
import pytest


def test_same_token(test_util):
    """
    Tests that we aren't immediately regenerating the key after each request
    """
    private_key, _ = test_util.generate_key_pair()
    sec_man = SecurityManager("testaccount", "snowman", private_key,
                              renewal_delay=timedelta(seconds=3))
    assert sec_man.get_token() == sec_man.get_token()


def test_regenerate_token(test_util):
    """
    Tests that the security manager generates new tokens after we
    cross the set renewal threshold
    """
    private_key, _ = test_util.generate_key_pair()
    sec_man = SecurityManager("testaccount", "snowman", private_key,
                              renewal_delay=timedelta(seconds=3))
    old_token = sec_man.get_token()
    sleep(5)
    assert old_token != sec_man.get_token()


def test_calculate_public_key_fingerprint(test_util):
    with open(os.path.join(test_util.get_data_dir(), 'test_rsa_key'), 'r') as key_file:
        private_key = key_file.read()
        sec_man = SecurityManager("testaccount", "snowman", private_key,
                                  renewal_delay=timedelta(minutes=3))
        public_key_fingerprint = sec_man.calculate_public_key_fingerprint(private_key)

        assert public_key_fingerprint == 'SHA256:QKX8hnXHVAVXp7mLdCAF+vjU2A8RBuRSpgdRjPHhVWY='


def test_invalid_private_key():
    sec_man = SecurityManager("testaccount", "snowman", 'invalid_private_key',
                              renewal_delay=timedelta(minutes=3))
    with pytest.raises(IngestClientError) as client_error:
        sec_man.get_token()

    assert client_error.value.code == ERR_INVALID_PRIVATE_KEY



# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.

"""
tokentools.py - provides services for automatically creating and renewing
JWT tokens for authenticating to Snowflake
"""

from datetime import timedelta, datetime
from logging import getLogger
from cryptography.exceptions import UnsupportedAlgorithm
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import PublicFormat
from cryptography.hazmat.backends import default_backend
from snowflake.connector.util_text import parse_account
from ..error import IngestClientError
from ..errorcode import ERR_INVALID_PRIVATE_KEY
import base64
import hashlib

import jwt

logger = getLogger(__name__)

try:
    from typing import Text
except ImportError:
    logger.debug('# Python 3.5.0 and 3.5.1 have incompatible typing modules.', exc_info=True)
    from typing_extensions import Text

ISSUER = "iss"
EXPIRE_TIME = "exp"
ISSUE_TIME = "iat"
SUBJECT = "sub"


class SecurityManager(object):
    """
    Given a private key, username, and account, signs and creates tokens for use in Snowflake Ingest Requests
    """
    LIFETIME = timedelta(minutes=59)  # The tokens will have a 59 minute lifetime
    RENEWAL_DELTA = timedelta(minutes=54)  # Tokens will be renewed after 54 minutes
    ALGORITHM = "RS256"  # Tokens will be generated using RSA with SHA256

    def __init__(self, account: Text, user: Text, private_key: Text,
                 lifetime: timedelta = LIFETIME, renewal_delay: timedelta = RENEWAL_DELTA):
        """
        __init__ creates a security manager with the specified context arguments
        :param account: the account in which data is being loaded
        :param user: The user who is loading these files
        :param private_key: the private key we'll use for signing tokens
        :param lifetime: how long this key will live (in minutes)
        :param renewal_delay: how long until the security manager should renew the key
        """

        logger.info(
            """Creating Security Manager with arguments
            account : %s, user : %s, lifetime : %s, renewal_delay : %s""",
            account, user, lifetime, renewal_delay)

        # Snowflake account names are canonically in all caps. Also trim account names if contain '.'
        self.account = parse_account(account.upper())
        self.user = user.upper()  # Snowflake user names are also in all caps by default
        self.qualified_username = self.account + "." + self.user  # Generate the full user name
        self.lifetime = lifetime  # the timedelta until our tokens expire
        self.renewal_delay = renewal_delay  # the timedelta until we renew the token
        self.private_key = private_key  # stash the private key
        self.renew_time = datetime.utcnow()  # We need to renew the token NOW
        self.token = None  # We initially have no token

    def get_token(self) -> Text:
        """
        Regenerates the current token if and only if we have exceeded the renewal time
        bounds set
        :return: the new token
        """
        now = datetime.utcnow()  # Fetch the current time

        # If the token has expired, or doesn't exist, regenerate it
        if self.token is None or self.renew_time <= now:
            logger.info("Renewing token because renewal time (%s) is eclipsed by present time (%s)",
                        self.renew_time, now)
            # Calculate the next time we need to renew the token
            self.renew_time = now + self.renewal_delay

            public_key_fp = self.calculate_public_key_fingerprint(self.private_key)

            # Create our payload
            payload = {

                # The issuer is the public key fingerprint
                ISSUER: self.qualified_username + '.' + public_key_fp,

                # subject is user's fully qualified username
                SUBJECT: self.qualified_username,

                # The payload was issued at this point it time
                ISSUE_TIME: now,

                # The token should no longer be accepted after our lifetime has elapsed
                EXPIRE_TIME: now + SecurityManager.LIFETIME
            }

            # Regenerate the actual token
            self.token = jwt.encode(payload, self.private_key, algorithm=SecurityManager.ALGORITHM)
            logger.info("New Token created")

        return self.token.decode('utf-8') if isinstance(self.token, bytes) else self.token

    def calculate_public_key_fingerprint(self, private_key: Text) -> Text:
        """
        Given a private key in pem format, return the public key fingerprint
        :param private_key: private key string
        :return: public key fingerprint
        """
        try:
            private_key = load_pem_private_key(private_key.encode(), None, default_backend())
        except (ValueError,  UnsupportedAlgorithm) as e:
            raise IngestClientError(
                    code=ERR_INVALID_PRIVATE_KEY,
                    message='Invalid private key. {}'.format(e))

        # get the raw bytes of public key
        public_key_raw = private_key.public_key().public_bytes(Encoding.DER, PublicFormat.SubjectPublicKeyInfo)

        # take sha256 on raw bytes and then do base64 encode
        sha256hash = hashlib.sha256()
        sha256hash.update(public_key_raw)

        public_key_fp = 'SHA256:' + base64.b64encode(sha256hash.digest()).decode('utf-8')
        logger.info("Public key fingerprint is %s", public_key_fp)

        return public_key_fp

    def get_account(self) -> Text:
        return self.account



from snowflake.ingest import SimpleIngestManager
from snowflake.ingest import StagedFile
import time
import os


def test_simple_ingest(connection_ctx, test_util):
    param = connection_ctx['param']

    pipe_name = '{}.{}.TEST_SIMPLE_INGEST_PIPE'.format(
        param['database'],
        param['schema'])

    private_key = test_util.read_private_key()

    cur = connection_ctx['cnx'].cursor()

    test_file = os.path.join(test_util.get_data_dir(), 'test_file.csv')
    cur.execute('create or replace table TEST_SIMPLE_INGEST_TABLE(c1 number, c2 string)')
    cur.execute('create or replace stage TEST_SIMPLE_INGEST_STAGE')
    cur.execute('put file://{} @TEST_SIMPLE_INGEST_STAGE'.format(test_file))
    cur.execute('create or replace pipe {0} as copy into TEST_SIMPLE_INGEST_TABLE '
                'from @TEST_SIMPLE_INGEST_STAGE'.format(pipe_name))

    ingest_manager = SimpleIngestManager(account=param['account'],
                                         user=param['user'],
                                         private_key=private_key,
                                         pipe=pipe_name,
                                         scheme=param['protocol'],
                                         host=param['host'],
                                         port=param['port'])

    staged_files = [StagedFile('test_file.csv.gz', None)]

    resp = ingest_manager.ingest_files(staged_files)

    assert resp['responseCode'] == 'SUCCESS'

    start_polling_time = time.time()

    while time.time() - start_polling_time < 120:
        history_resp = ingest_manager.get_history()

        if len(history_resp['files']) == 1:
            assert history_resp['files'][0]['path'] == 'test_file.csv.gz'
            return
        else:
            # wait for 20 seconds
            time.sleep(20)

    assert False




# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
from requests import Response

class IngestResponseError(Exception):
    """
        Error thrown when rest request failed
    """
    def __init__(self, response: Response):
        self.http_error_code = response.status_code

        try:
            json_body = response.json()
        except ValueError:
            self.message = 'Http Error: {}, Message: {}'.format(self.http_error_code,
                                                                response.reason)
            return

        try:
            self.code = json_body[u'code']
            self.success = json_body[u'success']
            self._raw_message = json_body[u'message']
            self.data = json_body[u'data']
            self.message = 'Http Error: {}, Vender Code: {}, Message: {}' \
                .format(self.http_error_code, self.code, self._raw_message)
        except KeyError:
            self.message = 'Http Error: {}, Message: {}, Body: {}'.format(self.http_error_code,
                                                                          response.reason,
                                                                          json_body)
        return

    def __str__(self):
        return self.message


class IngestClientError(Exception):
    """
        Error thrown in the client side
    """
    def __init__(self, **kwargs):
        self.code = kwargs.get('code')
        self.message = kwargs.get('message')

    def __str__(self):
        return 'Vendor Code: {}, Message: {}'\
            .format(self.code, self.message)
